from kiwipiepy import Kiwi


def fetch_kiwi() -> Kiwi:
    """
    fetch kiwi with user-defined rules
    """
    kiwi = Kiwi()
    kiwi.add_user_word(".", tag="SF")
    kiwi.add_user_word("우리말", tag="NNP", score=10)
    kiwi.add_user_word("점순이", tag="NNP", score=10)
    kiwi.add_user_word("란다", tag="EF", score=10)
    kiwi.add_user_word("래요", tag="EF", score=10)
    kiwi.add_user_word("ㄴ가요", tag="EF", score=10)
    kiwi.add_user_word("랍니다", tag="EF", score=10)
    kiwi.add_user_word("할금할금", tag="MAG", score=10)
    kiwi.add_pre_analyzed_word(
        "벗어.", [("벗", "VV-R"), ("어", "EF"), (".", "SF")], score=10
    )
    kiwi.add_pre_analyzed_word(
        "누구신가요", [("누구", "NP"), ("이", "VCP"), ("시", "EP"), ("ᆫ가요", "EF")], score=10
    )
    return kiwi



import re
from typing import List
from politely.rules import RULES


class SFNotIncludedError(Exception):
    """
    An exception raised when a sentence does not include a SF.
    """

    def __init__(self, out: str):
        self.out = out

    def __str__(self) -> str:
        return "Sentence does not include a SF:\n" + "\n".join(self.out)


class EFNotIncludedError(Exception):
    """
    An exception raised when a sentence does not include an EF.
    """

    def __init__(self, out: str):
        self.out = out

    def __str__(self) -> str:
        return "Sentence does not include an EF:\n" + "\n".join(self.out)


class EFNotSupportedError(Exception):
    """
    An exception raised when a sentence includes an EF that is not supported.
    """
    def __init__(self, out: str):
        self.out = out

    def __str__(self) -> str:
        return "EF is not supported:\n" + "\n".join(self.out)



"""
규칙:
1.
"""
from typing import Set, Tuple, Dict

# --- symbols --- #
NULL = "❌"
TAG = "🏷"
SEP = "🔗"


# --- all EF's of different styles they must be singular tokens --- #
CASUAL = {
    f"어{TAG}EF",
    f"다{TAG}EF",
    f"니{TAG}EF",
    f"라{TAG}EF",
    f"어라{TAG}EF",
    f"자{TAG}EF",
    f"대{TAG}EF",
    f"는다{TAG}EF",
    f"마{TAG}EF",
    f"야{TAG}EF",
    f"군{TAG}EF",
    f"네{TAG}EF",
    f"냐{TAG}EF",
    f"ᆫ다{TAG}EF",
    f"란다{TAG}EF",
    f"ᆯ게{TAG}EF",
    f"ᆫ대{TAG}EF",
    f"ᆫ가{TAG}EF",
    f"지{TAG}EF"
}

POLITE = {
    f"요{TAG}EF",
    f"어요{TAG}EF",
    f"에요{TAG}EF",
    f"ᆫ가요{TAG}EF",
    f"지요{TAG}EF",
    f"래요{TAG}EF",
    f"죠{TAG}EF",
    f"래요{TAG}EF",
    f"네요{TAG}EF",
    f"나요{TAG}EF",
    f"대요{TAG}EF",
    f"ᆯ게요{TAG}EF",
    f"ᆫ대요{TAG}EF",
    f"ᆫ가요{TAG}EF",
    f"세요{TAG}EF"
}

FORMAL = {
    f"습니다{TAG}EF",
    f"습니까{TAG}EF",
    f"랍니다{TAG}EF",
    f"ᆸ니까{TAG}EF",
    f"ᆸ시오{TAG}EF",
    f"ᆸ니다{TAG}EF",
    f"ᆸ시다{TAG}EF",
    f"읍시다{TAG}EF",
}


# --- regex --- #
EFS = rf"(?P<MASK>({'|'.join([pair for pair in (CASUAL | POLITE | FORMAL)])}))"
SELF = rf"\g<MASK>"
WITH_JONG_SUNG = rf"[{''.join({chr(i) for i in range(44032, 55204)} - {chr(44032 + 28 * i) for i in range(399)})}]"
NO_JONG_SUNG = rf"[^{''.join({chr(i) for i in range(44032, 55204)} - {chr(44032 + 28 * i) for i in range(399)})}]"


# --- programmatically populated RULES --- #
RULES: Dict[str, Tuple[Set[str], Set[str], Set[str]]] = dict()

# --- the overarching rule --- #
RULES.update({
    EFS: (
        CASUAL,
        POLITE,
        FORMAL
    )
})

# --- 시/EP: 이미 시/EP가 존재하는 경우, 반말을 쓸 때 제거한다 --- #
RULES.update(
    {
        rf"(?P<MASK>(시|으시){TAG}EP)": (
            {NULL},  # you don't use them
            {SELF},  # just repeat yourself
            {SELF},  # just repeat yourself
        )
    }
)


# --- 종성이 있는 경우, 종성으로 시작하는 EF는 사용하지 않음 --- #
RULES.update(
    {
        rf"{WITH_JONG_SUNG}{TAG}[A-Z\-]+?{SEP}{EFS}": (
            CASUAL - {f"ᆫ다{TAG}EF", f"ᆯ게{TAG}EF", f"ᆫ대{TAG}EF"},
            POLITE - {f"ᆯ게요{TAG}EF", f"ᆫ대요{TAG}EF", f"ᆫ가요{TAG}EF"},
            FORMAL - {f"ᆸ니까{TAG}EF", f"ᆸ시오{TAG}EF", f"ᆸ니다{TAG}EF", f"ᆸ시다{TAG}EF"}
        )
    }
)

# --- 종성이 없는 경우, 어/EF, f"어라{TAG}EF", 어요/EF는 사용하지 않음 --- #
RULES.update(
    {
        rf"{NO_JONG_SUNG}{TAG}[A-Z\-]+?{SEP}{EFS}": (
            CASUAL - {f"어{TAG}EF", f"어라{TAG}EF"},
            POLITE - {f"어요{TAG}EF"},
            FORMAL
        )
    }
)


# --- 의문형인 경우, formal은 -니까만 가능 --- #
RULES.update(
    {
        rf"{EFS}{SEP}\?{TAG}SF": (
            CASUAL,
            POLITE,
            {f"습니까{TAG}EF", f"ᆸ니까{TAG}EF", f"시{TAG}EP{SEP}ᆸ니까{TAG}EF"}
        )
    }
)

# --- 나/저 --- #
RULES.update(
    {
        rf"(?P<MASK>(나|저){TAG}NP)": (
            {f"나{TAG}NP"},
            {f"저{TAG}NP"},
            {f"저{TAG}NP"}
        )
    }
)


# --- 너/당신 --- #
RULES.update(
    {
        rf"(?P<MASK>(너|당신){TAG}NP)": (
            {f"너{TAG}NP"},
            {f"당신{TAG}NP"},
            {f"당신{TAG}NP"}
        )
    }
)


# --- 엄마/어머니 --- #
RULES.update(  # noqa
    {
        rf"(?P<MASK>(엄마|어머니|어머님){TAG}NNG)": (
            {f"엄마{TAG}NNG"},
            {f"어머니{TAG}NNG", f"어머님{TAG}NNG"},
            {f"어머니{TAG}NNG", f"어머님{TAG}NNG"}
        )
    }
)


# --- 아빠/아버지 --- #
RULES.update(
    {
        rf"(?P<MASK>(아빠|아버지|아버님){TAG}NNG)": (
            {f"아빠{TAG}NNG"},
            {f"아버지{TAG}NNG", f"아버님{TAG}NNG"},
            {f"아버지{TAG}NNG", f"아버님{TAG}NNG"}
        )
    }
)

# --- 께서 --- #
RULES.update(
    {
        rf"(어머니|어머님|아버지|아버님|선생님|할머니|할아버지){TAG}NNG{SEP}(?P<MASK>\S+?{TAG}JKS)": (
            {SELF},
            {f"께서{TAG}JKS"},
            {f"께서{TAG}JKS"}
        )
    }
)

# --- 란다 -> 래요 / 랍니다 --- #
RULES.update(
    {
        rf"(?P<MASK>(란다|래요|랍니다){TAG}EF)": (
            {f"란다{TAG}EF"},
            {f"래요{TAG}EF"},
            {f"랍니다{TAG}EF"}
        )
    }
)


# --- 지 -> 지요 --- #
RULES.update(
    {
        rf"(?P<MASK>(지|지요|ᆸ니다){TAG}EF)": (
            {f"지{TAG}EF"},
            {f"지요{TAG}EF"},
            #  전부 가능함
            FORMAL
        )
    }
)


# --- 시 + ㄴ가요 -> 시 + 니 --- #
RULES.update(
    {
        rf"시{TAG}EP{SEP}(?P<MASK>ᆫ가요{TAG}EF)": (
            {f"니{TAG}EF"},
            {f"ᆫ가요{TAG}EF"},
            {f"ᆸ니까{TAG}EF"}
        )
    }
)


# # --- 으세요 -> 으십시오 --- #
# RULES.update(
#     {
#         rf"(?P<MASK>으세요{TAG}EF)": (
#             {f"어{TAG}EF"},
#             {f"으세요{TAG}EF"},
#             {f"ᆸ시오{TAG}EF"}
#         )
#     }
# )

# ---- to be used for scoring -- #
PREFERENCES = {
    f"어{TAG}EF",
    f"어요{TAG}EF",
    f"어요{TAG}EF",
    f"습니다{TAG}EF",
    f"ᆸ니다{TAG}EF"
}


import itertools
import os
import re
from copy import copy, deepcopy
import random
from typing import Any, Tuple, Dict, Set
from functools import wraps
import numpy as np
import torch
from politely.errors import EFNotSupportedError, SFNotIncludedError, EFNotIncludedError
from politely.fetchers import fetch_kiwi
from politely import RULES, SEP, TAG, NULL, SELF
from politely.modeling_gpt2_scorer import GPT2Scorer
from politely.modeling_heuristic_scorer import HeuristicScorer
from politely.modeling_scorer import Scorer
from politely.rules import EFS


def log(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        # get the function signature
        f_out = f(*args, **kwargs)
        names = f.__code__.co_varnames[: f.__code__.co_argcount]
        styler_instance: Styler = args[0]
        # exclude self
        styler_instance.log[f.__name__] = {"in": dict(zip(names[1:], args[1:])), "out": copy(styler_instance.out)}
        return f_out # return the out
    return wrapper


class Styler:
    """
    A rule-based Korean Politeness Styler
    """
    def __init__(self, strict: bool = False, scorer: str = "gpt2"):
        #  --- object-owned attributes --- #\
        if scorer == "heuristic":
            self.scorer: Scorer = HeuristicScorer()
        elif scorer == "gpt2":
            self.scorer: Scorer = GPT2Scorer()
        else:
            raise ValueError(f"scorer should be either 'heuristic' or 'gpt2', but got {scorer}")
        self.strict = strict
        self.out: Any = None
        self.kiwi = fetch_kiwi()
        self.rules = deepcopy(RULES)
        self.log = dict()

    def __call__(self, sent: str, politeness: int) -> str:
        """
        Style a sentence with the given politeness (0, 1, 2)
        """
        self.setup() \
            .preprocess(sent) \
            .analyze() \
            .check() \
            .honorify(politeness) \
            .guess() \
            .elect() \
            .conjugate()
        return self.out

    def setup(self):
        """
        Reset the out and clear all the logs,
        """
        self.out = None
        self.log.clear()
        self.log.update({"conjugations": set(), "honorifics": set()})
        # --- seed everything --- #
        seed = 318
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = True
        return self

    @log
    def preprocess(self, sent: str):
        """
        Make sure each sentence ends with a period, if it does not end with any SF.
        We do this to increase the accuracy of `kiwi.join`.
        """
        self.out = re.sub(r"([^!?.]+)$", r"\1.", sent.strip())
        return self

    @log
    def analyze(self):
        """
        Analyze the sentence and generate the output.
        """
        self.out: str
        self.out = [f"{token.form}{TAG}{token.tag}" for token in self.kiwi.tokenize(self.out)]
        self.out = SEP.join(self.out)  # to match with the format of the rules
        return self

    def check(self):
        """
        Check if your assumption holds. Raises a custom error if any of them does not hold.
        """
        self.out: str
        # raise exceptions only if you are in debug mode
        if self.strict:
            # assumption 1: every sentence should end with a valid SF. It should be one of: (., !, ?)
            if "SF" not in self.out:
                raise SFNotIncludedError(self.out)
            # assumption 2: every sentence should include a valid EF. It should match the EFS pattern.
            if "EF" not in self.out:
                raise EFNotIncludedError(self.out)
            # assumption 3: all EF's should be supported by politely.
            if not all([re.match(EFS, pair) for pair in self.out.split(SEP) if "EF" in pair]):
                raise EFNotSupportedError(self.out)
        return self

    @log
    def honorify(self, politeness: int):
        """
        Determines all the candidates that would properly honorify the sentence.
        Do this by chain-conjugating sets.
        """
        self.out: str
        pair2honorifics = {}
        for pattern in self.rules.keys():
            match = re.search(pattern, self.out)
            if match:
                # should be only one pair
                pair = match.group("MASK")
                honorifics = {honorific.replace(SELF, pair) for honorific in self.rules[pattern][politeness]}
                # progressively narrow down honorifics
                pair2honorifics[pair] = pair2honorifics.get(pair, honorifics) & honorifics
        # get all possible candidates
        candidates = itertools.product(*[
            pair2honorifics.get(pair, {pair, })
            for pair in self.out.split(SEP)  # SEP
        ])
        # remove empty candidates
        candidates = [
            [pair.split(SEP) for pair in candidate if pair != NULL]
            for candidate in candidates
        ]
        # flatten pairs
        candidates = [
            list(itertools.chain(*candidate))
            for candidate in candidates
        ]
        # a list of candidates
        self.out = candidates
        return self

    @log
    def guess(self):
        """
        Guess the scores.
        """
        self.out: list[list[str]]
        scores = self.scorer(self.out, self.log, self.kiwi)
        self.out = [(candidate, score) for candidate, score in zip(self.out, scores)]
        return self
    
    @log
    def elect(self):
        """
        Elect the best candidate.
        """
        self.out: list[tuple[list[str], float]]
        best = max(self.out, key=lambda x: x[1])
        self.out = best
        return self

    @log
    def conjugate(self):
        """
        conjugate the best candidate.
        """
        self.out: tuple[list[str], float]
        sent = self.out[0]
        self.out = self.kiwi.join([(pair.split(TAG)[0], pair.split(TAG)[1]) for pair in sent])
        return self

    def add_rules(self, rules: Dict[str, Tuple[Set,
                                               Set,
                                               Set]]):
        """
        Add rules to the existing rules.
        """
        # check if the rules are in proper format
        for key, _ in rules.items():
            # first, check if the key includes a group with the key; (?<MASK>...)
            # e.g. (?P<MASK>(아빠|아버지){TAG}NNG)
            if not re.search(re.escape(r"(?P<MASK>") + r".*" + re.escape(")"), key):
                raise ValueError(f"key should include a group with the key; (?P<MASK>...), but got {key}")
        self.rules.update(rules)
        return self
        


from typing import List
import torch
from kiwipiepy import Kiwi
from torch.nn import CrossEntropyLoss
from transformers import AutoTokenizer, GPT2LMHeadModel
from politely import TAG
from politely.modeling_scorer import Scorer


class GPT2Scorer(Scorer):
    """
    More accurate than heuristic scorer, but slower.
    """

    def __init__(
        self, device: str = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    ):
        self.gpt2 = GPT2LMHeadModel.from_pretrained(
            "beomi/kykim-gpt3-kor-small_based_on_gpt2"
        ).to(device)
        self.tokenizer = AutoTokenizer.from_pretrained(
            "beomi/kykim-gpt3-kor-small_based_on_gpt2"
        )
        self.device = device
        self.gpt2.eval()

    def __call__(
        self,
        candidates: List[List[str]],
        logs: dict,
        kiwi: Kiwi,
        # label_smoothing: https://ratsgo.github.io/insight-notes/docs/interpretable/smoothing
        label_smoothing: float = 0.2,
    ) -> list[float]:
        """
        # label_smoothing: https://ratsgo.github.io/insight-notes/docs/interpretable/smoothing
        이걸 batched processing으로 바꾸고 계속하기. device parameter도 추가하기.
        compute the negative log likelihood of the candidates.
        """
        # first, join the pairs into a sentence
        sents = [
            kiwi.join(tuple(pair.split(TAG)) for pair in candidate)
            for candidate in candidates
        ]
        inputs = self.tokenizer(
            sents, return_tensors="pt", padding=True, truncation=True
        ).to(
            self.device
        )  # (N, L)
        with torch.no_grad():
            # output the loss of each instance
            outputs = self.gpt2(**inputs)
        logits = outputs["logits"]  # (N, L, |V|)
        # --- shift them --- #
        shift_logits = logits[:, :-1, :].contiguous()  # (N, L-1, |V|)
        shift_labels = inputs.input_ids[:, 1:].contiguous()  # (N, L-1)
        # flatten the tokens
        # TODO: use weight parameter to increase the losses for canonical ending like -습니다 / -어 / -어요
        loss_fct = CrossEntropyLoss(
            reduction="none",
            ignore_index=self.tokenizer.pad_token_id,
            label_smoothing=label_smoothing,
        )
        loss = loss_fct(
            shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1)
        )
        loss_by_sequence = loss.view(logits.size(0), -1).mean(dim=1)
        # negative log likelihood
        loss_by_sequence = -loss_by_sequence
        return loss_by_sequence.tolist()



from typing import List
from kiwipiepy import Kiwi
from politely import PREFERENCES, CASUAL, POLITE, FORMAL, SEP
from politely.modeling_scorer import Scorer


class HeuristicScorer(Scorer):

    def __call__(self, candidates: List[List[str]], logs: dict, kiwi: Kiwi) -> List[List[float]]:
        """
        A naive scoring strategy that relies on the heuristic rules.
        """
        politeness = logs['honorify']['in']['politeness']
        original_pairs = logs['analyze']['out'].split(SEP)
        if politeness == 0:
            boost_pairs = CASUAL & set(original_pairs)
        elif politeness == 1:
            boost_pairs = POLITE & set(original_pairs)
        elif politeness == 2:
            boost_pairs = FORMAL & set(original_pairs)
        else:
            raise ValueError(f"politeness should be one of (0, 1, 2), but got {politeness}")
        scores = list()
        for pairs in candidates:
            # if it includes priority pairs, it's a good candidate (at least for this scorer)
            pairs = set(pairs)
            score = (len(PREFERENCES & pairs) / len(pairs)) * 0.1
            score += (len(boost_pairs & pairs) / len(pairs)) * 0.9
            scores.append(score)
        return scores
